@@ -1,6 +1,8 @@
+/*	$NetBSD: exec.c,v 1.24 1997/07/20 21:27:36 christos Exp $	*/
+
 /*-
- * Copyright (c) 1991 The Regents of the University of California.
- * All rights reserved.
+ * Copyright (c) 1991, 1993
+ *	The Regents of the University of California.  All rights reserved.
  *
  * This code is derived from software contributed to Berkeley by
  * Kenneth Almquist.
@@ -34,11 +36,22 @@
  * SUCH DAMAGE.
  */
 
+#include <sys/cdefs.h>
 #ifndef lint
-/*static char sccsid[] = "from: @(#)exec.c	5.2 (Berkeley) 3/13/91";*/
-static char rcsid[] = "exec.c,v 1.5 1993/08/01 18:58:17 mycroft Exp";
+#if 0
+static char sccsid[] = "@(#)exec.c	8.4 (Berkeley) 6/8/95";
+#else
+__RCSID("$NetBSD: exec.c,v 1.24 1997/07/20 21:27:36 christos Exp $");
+#endif
 #endif /* not lint */
 
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <unistd.h>
+#include <fcntl.h>
+#include <errno.h>
+#include <stdlib.h>
+
 /*
  * When commands are first encountered, they are entered in a hash table.
  * This ensures that a full path search will not have to be done for them
@@ -65,15 +78,9 @@ static char rcsid[] = "exec.c,v 1.5 1993/08/01 18:58:17 mycroft Exp";
 #include "error.h"
 #include "init.h"
 #include "mystring.h"
-#include <sys/types.h>
-#include <sys/stat.h>
-#include <fcntl.h>
-#include <errno.h>
-#ifdef  BSD
-#undef BSD      /* temporary, already defined in <sys/param.h> */
-#include <sys/param.h>
-#include <unistd.h>
-#endif
+#include "show.h"
+#include "jobs.h"
+#include "alias.h"
 
 
 #define CMDTABLESIZE 31		/* should be prime */
@@ -92,23 +99,15 @@ struct tblentry {
 
 STATIC struct tblentry *cmdtable[CMDTABLESIZE];
 STATIC int builtinloc = -1;		/* index in path of %builtin, or -1 */
+int exerrno = 0;			/* Last exec error */
 
 
-#ifdef __STDC__
-STATIC void tryexec(char *, char **, char **);
-STATIC void execinterp(char **, char **);
-STATIC void printentry(struct tblentry *);
-STATIC void clearcmdentry(int);
-STATIC struct tblentry *cmdlookup(char *, int);
-STATIC void delete_cmd_entry(void);
-#else
-STATIC void tryexec();
-STATIC void execinterp();
-STATIC void printentry();
-STATIC void clearcmdentry();
-STATIC struct tblentry *cmdlookup();
-STATIC void delete_cmd_entry();
-#endif
+STATIC void tryexec __P((char *, char **, char **));
+STATIC void execinterp __P((char **, char **));
+STATIC void printentry __P((struct tblentry *, int));
+STATIC void clearcmdentry __P((int));
+STATIC struct tblentry *cmdlookup __P((char *, int));
+STATIC void delete_cmd_entry __P((void));
 
 
 
@@ -121,7 +120,8 @@ void
 shellexec(argv, envp, path, index)
 	char **argv, **envp;
 	char *path;
-	{
+	int index;
+{
 	char *cmdname;
 	int e;
 
@@ -139,7 +139,20 @@ shellexec(argv, envp, path, index)
 			stunalloc(cmdname);
 		}
 	}
-	error2(argv[0], errmsg(e, E_EXEC));
+
+	/* Map to POSIX errors */
+	switch (e) {
+	case EACCES:
+		exerrno = 126;
+		break;
+	case ENOENT:
+		exerrno = 127;
+		break;
+	default:
+		exerrno = 2;
+		break;
+	}
+	exerror(EXEXEC, "%s: %s", argv[0], errmsg(e, E_EXEC));
 }
 
 
@@ -150,7 +163,9 @@ tryexec(cmd, argv, envp)
 	char **envp;
 	{
 	int e;
+#ifndef BSD
 	char *p;
+#endif
 
 #ifdef SYSV
 	do {
@@ -276,7 +291,7 @@ padvance(path, name)
 	char **path;
 	char *name;
 	{
-	register char *p, *q;
+	char *p, *q;
 	char *start;
 	int len;
 
@@ -289,7 +304,7 @@ padvance(path, name)
 		growstackblock();
 	q = stackblock();
 	if (p != start) {
-		bcopy(start, q, p - start);
+		memcpy(q, start, p - start);
 		q += p - start;
 		*q++ = '/';
 	}
@@ -311,7 +326,11 @@ padvance(path, name)
 /*** Command hashing code ***/
 
 
-hashcmd(argc, argv)  char **argv; {
+int
+hashcmd(argc, argv)
+	int argc;
+	char **argv;
+{
 	struct tblentry **pp;
 	struct tblentry *cmdp;
 	int c;
@@ -319,14 +338,6 @@ hashcmd(argc, argv)  char **argv; {
 	struct cmdentry entry;
 	char *name;
 
-	if (argc <= 1) {
-		for (pp = cmdtable ; pp < &cmdtable[CMDTABLESIZE] ; pp++) {
-			for (cmdp = *pp ; cmdp ; cmdp = cmdp->next) {
-				printentry(cmdp);
-			}
-		}
-		return 0;
-	}
 	verbose = 0;
 	while ((c = nextopt("rv")) != '\0') {
 		if (c == 'r') {
@@ -335,16 +346,24 @@ hashcmd(argc, argv)  char **argv; {
 			verbose++;
 		}
 	}
+	if (*argptr == NULL) {
+		for (pp = cmdtable ; pp < &cmdtable[CMDTABLESIZE] ; pp++) {
+			for (cmdp = *pp ; cmdp ; cmdp = cmdp->next) {
+				printentry(cmdp, verbose);
+			}
+		}
+		return 0;
+	}
 	while ((name = *argptr) != NULL) {
 		if ((cmdp = cmdlookup(name, 0)) != NULL
 		 && (cmdp->cmdtype == CMDNORMAL
-		     || cmdp->cmdtype == CMDBUILTIN && builtinloc >= 0))
+		     || (cmdp->cmdtype == CMDBUILTIN && builtinloc >= 0)))
 			delete_cmd_entry();
-		find_command(name, &entry, 1);
+		find_command(name, &entry, DO_ERR, pathval());
 		if (verbose) {
 			if (entry.cmdtype != CMDUNKNOWN) {	/* if no error msg */
 				cmdp = cmdlookup(name, 0);
-				printentry(cmdp);
+				printentry(cmdp, verbose);
 			}
 			flushall();
 		}
@@ -355,8 +374,9 @@ hashcmd(argc, argv)  char **argv; {
 
 
 STATIC void
-printentry(cmdp)
+printentry(cmdp, verbose)
 	struct tblentry *cmdp;
+	int verbose;
 	{
 	int index;
 	char *path;
@@ -374,6 +394,14 @@ printentry(cmdp)
 		out1fmt("builtin %s", cmdp->cmdname);
 	} else if (cmdp->cmdtype == CMDFUNCTION) {
 		out1fmt("function %s", cmdp->cmdname);
+		if (verbose) {
+			INTOFF;
+			name = commandtext(cmdp->param.func);
+			out1c(' ');
+			out1str(name);
+			ckfree(name);
+			INTON;
+		}
 #ifdef DEBUG
 	} else {
 		error("internal error: cmdtype %d", cmdp->cmdtype);
@@ -392,14 +420,15 @@ printentry(cmdp)
  */
 
 void
-find_command(name, entry, printerr)
+find_command(name, entry, act, path)
 	char *name;
 	struct cmdentry *entry;
-	{
+	int act;
+	char *path;
+{
 	struct tblentry *cmdp;
 	int index;
 	int prev;
-	char *path;
 	char *fullname;
 	struct stat statb;
 	int e;
@@ -407,6 +436,22 @@ find_command(name, entry, printerr)
 
 	/* If name contains a slash, don't use the hash table */
 	if (strchr(name, '/') != NULL) {
+		if (act & DO_ABS) {
+			while (stat(name, &statb) < 0) {
+	#ifdef SYSV
+				if (errno == EINTR)
+					continue;
+	#endif
+				if (errno != ENOENT && errno != ENOTDIR)
+					e = errno;
+				entry->cmdtype = CMDUNKNOWN;
+				entry->u.index = -1;
+				return;
+			}
+			entry->cmdtype = CMDNORMAL;
+			entry->u.index = -1;
+			return;
+		}
 		entry->cmdtype = CMDNORMAL;
 		entry->u.index = 0;
 		return;
@@ -435,7 +480,6 @@ find_command(name, entry, printerr)
 			prev = cmdp->param.index;
 	}
 
-	path = pathval();
 	e = ENOENT;
 	index = -1;
 loop:
@@ -475,7 +519,7 @@ loop:
 			goto loop;
 		}
 		e = EACCES;	/* if we fail, this will be the error */
-		if ((statb.st_mode & S_IFMT) != S_IFREG)
+		if (!S_ISREG(statb.st_mode))
 			goto loop;
 		if (pathopt) {		/* this is a %func directory */
 			stalloc(strlen(fullname) + 1);
@@ -485,36 +529,17 @@ loop:
 			stunalloc(fullname);
 			goto success;
 		}
-		if (geteuid() == 0) {
-			if ((statb.st_mode & 0111) == 0)
-				goto loop;
-		} else if (statb.st_uid == geteuid()) {
+#ifdef notdef
+		if (statb.st_uid == geteuid()) {
 			if ((statb.st_mode & 0100) == 0)
 				goto loop;
 		} else if (statb.st_gid == getegid()) {
 			if ((statb.st_mode & 010) == 0)
 				goto loop;
 		} else {
-			if ((statb.st_mode & 01) == 0) {
-#ifdef  BSD
-				if ((statb.st_mode & 010) == 0)
-					goto loop;
-				/* Are you in this group too? */
-				{
-					int group_list[NGROUPS];
-					int ngroups, i;
-
-					ngroups = getgroups(NGROUPS, group_list);
-					for (i = 0; i < ngroups; i++)
-						if (statb.st_gid == group_list[i])
-							goto Found;
-				}
-#endif
+			if ((statb.st_mode & 01) == 0)
 				goto loop;
-			}
 		}
-#ifdef  BSD
-	Found:
 #endif
 		TRACE(("searchexec \"%s\" returns \"%s\"\n", name, fullname));
 		INTOFF;
@@ -528,7 +553,7 @@ loop:
 	/* We failed.  If there was an entry for this command, delete it */
 	if (cmdp)
 		delete_cmd_entry();
-	if (printerr)
+	if (act & DO_ERR)
 		outfmt(out2, "%s: %s\n", name, errmsg(e, E_EXEC));
 	entry->cmdtype = CMDUNKNOWN;
 	return;
@@ -548,8 +573,8 @@ success:
 int
 find_builtin(name)
 	char *name;
-	{
-	const register struct builtincmd *bp;
+{
+	const struct builtincmd *bp;
 
 	for (bp = builtincmd ; bp->name ; bp++) {
 		if (*bp->name == *name && equal(bp->name, name))
@@ -573,7 +598,7 @@ hashcd() {
 	for (pp = cmdtable ; pp < &cmdtable[CMDTABLESIZE] ; pp++) {
 		for (cmdp = *pp ; cmdp ; cmdp = cmdp->next) {
 			if (cmdp->cmdtype == CMDNORMAL
-			 || cmdp->cmdtype == CMDBUILTIN && builtinloc >= 0)
+			 || (cmdp->cmdtype == CMDBUILTIN && builtinloc >= 0))
 				cmdp->rehash = 1;
 		}
 	}
@@ -589,9 +614,9 @@ hashcd() {
 
 void
 changepath(newval)
-	char *newval;
-	{
-	char *old, *new;
+	const char *newval;
+{
+	const char *old, *new;
 	int index;
 	int firstchange;
 	int bltin;
@@ -604,8 +629,8 @@ changepath(newval)
 	for (;;) {
 		if (*old != *new) {
 			firstchange = index;
-			if (*old == '\0' && *new == ':'
-			 || *old == ':' && *new == '\0')
+			if ((*old == '\0' && *new == ':')
+			 || (*old == ':' && *new == '\0'))
 				firstchange++;
 			old = new;	/* ignore subsequent differences */
 		}
@@ -633,7 +658,9 @@ changepath(newval)
  */
 
 STATIC void
-clearcmdentry(firstchange) {
+clearcmdentry(firstchange)
+	int firstchange;
+{
 	struct tblentry **tblp;
 	struct tblentry **pp;
 	struct tblentry *cmdp;
@@ -642,8 +669,10 @@ clearcmdentry(firstchange) {
 	for (tblp = cmdtable ; tblp < &cmdtable[CMDTABLESIZE] ; tblp++) {
 		pp = tblp;
 		while ((cmdp = *pp) != NULL) {
-			if (cmdp->cmdtype == CMDNORMAL && cmdp->param.index >= firstchange
-			 || cmdp->cmdtype == CMDBUILTIN && builtinloc >= firstchange) {
+			if ((cmdp->cmdtype == CMDNORMAL &&
+			     cmdp->param.index >= firstchange)
+			 || (cmdp->cmdtype == CMDBUILTIN &&
+			     builtinloc >= firstchange)) {
 				*pp = cmdp->next;
 				ckfree(cmdp);
 			} else {
@@ -660,7 +689,7 @@ clearcmdentry(firstchange) {
  */
 
 #ifdef mkinit
-MKINIT void deletefuncs();
+MKINIT void deletefuncs __P((void));
 
 SHELLPROC {
 	deletefuncs();
@@ -705,9 +734,10 @@ struct tblentry **lastcmdentry;
 STATIC struct tblentry *
 cmdlookup(name, add)
 	char *name;
-	{
+	int add;
+{
 	int hashval;
-	register char *p;
+	char *p;
 	struct tblentry *cmdp;
 	struct tblentry **pp;
 
@@ -736,7 +766,6 @@ cmdlookup(name, add)
 	return cmdp;
 }
 
-
 /*
  * Delete the command entry returned on the last lookup.
  */
@@ -758,7 +787,7 @@ delete_cmd_entry() {
 void
 getcmdentry(name, entry)
 	char *name;
-	struct cmdentry *entry; 
+	struct cmdentry *entry;
 	{
 	struct tblentry *cmdp = cmdlookup(name, 0);
 
@@ -819,7 +848,7 @@ defun(name, func)
  * Delete a function if it exists.
  */
 
-void
+int
 unsetfunc(name)
 	char *name;
 	{
@@ -828,5 +857,85 @@ unsetfunc(name)
 	if ((cmdp = cmdlookup(name, 0)) != NULL && cmdp->cmdtype == CMDFUNCTION) {
 		freefunc(cmdp->param.func);
 		delete_cmd_entry();
+		return (0);
 	}
+	return (1);
+}
+
+/*
+ * Locate and print what a word is...
+ */
+
+int
+typecmd(argc, argv)
+	int argc;
+	char **argv;
+{
+	struct cmdentry entry;
+	struct tblentry *cmdp;
+	char **pp;
+	struct alias *ap;
+	int i;
+	int error = 0;
+	extern char *const parsekwd[];
+
+	for (i = 1; i < argc; i++) {
+		out1str(argv[i]);
+		/* First look at the keywords */
+		for (pp = (char **)parsekwd; *pp; pp++)
+			if (**pp == *argv[i] && equal(*pp, argv[i]))
+				break;
+
+		if (*pp) {
+			out1str(" is a shell keyword\n");
+			continue;
+		}
+
+		/* Then look at the aliases */
+		if ((ap = lookupalias(argv[i], 1)) != NULL) {
+			out1fmt(" is an alias for %s\n", ap->val);
+			continue;
+		}
+
+		/* Then check if it is a tracked alias */
+		if ((cmdp = cmdlookup(argv[i], 0)) != NULL) {
+			entry.cmdtype = cmdp->cmdtype;
+			entry.u = cmdp->param;
+		}
+		else {
+			/* Finally use brute force */
+			find_command(argv[i], &entry, DO_ABS, pathval());
+		}
+
+		switch (entry.cmdtype) {
+		case CMDNORMAL: {
+			int j = entry.u.index;
+			char *path = pathval(), *name;
+			if (j == -1) 
+				name = argv[i];
+			else {
+				do { 
+					name = padvance(&path, argv[i]);
+					stunalloc(name);
+				} while (--j >= 0);
+			}
+			out1fmt(" is%s %s\n",
+			    cmdp ? " a tracked alias for" : "", name);
+			break;
+		}
+		case CMDFUNCTION:
+			out1str(" is a shell function\n");
+			break;
+
+		case CMDBUILTIN:
+			out1str(" is a shell builtin\n");
+			break;
+
+		default:
+			out1str(" not found\n");
+			error |= 127;
+			break;
+		}
+	}
+	return error;
 }
